import sys
from pathlib import Path
from struct import unpack

DEBUG = False
if len(sys.argv) > 2:
    DEBUG = True

rom = Path(sys.argv[1]).read_bytes()[0x10:]
output_path = Path("extracted")
output_path.mkdir(exist_ok=True)

base = 0x1F8000
addrs_pos = 0x84BE - 0x8000
info_addrs_pos = 0x9CCE - 0x8000

headers = {
0xF1: ("10 00 70 00 00 00 00 00 00 00 00 00", 256,   0, "007"), # 7   , PRG 256, CHR RAM
0xF2: ("08 00 70 00 00 00 00 00 00 00 00 00", 128,   0, "007"), # 7   , PRG 128, CHR RAM
0x00: ("20 20 40 00 00 00 00 00 00 00 00 00", 512, 256, "004"), # MMC3, PRG 512, CHR ROM 256
0x01: ("10 20 40 00 00 00 00 00 00 00 00 00", 256, 256, "004"), # MMC3, PRG 256, CHR ROM 256
0x11: ("10 10 40 00 00 00 00 00 00 00 00 00", 256, 128, "004"), # MMC3, PRG 256, CHR ROM 128
0x02: ("08 20 40 00 00 00 00 00 00 00 00 00", 128, 256, "004"), # MMC3, PRG 128, CHR ROM 256
0x12: ("08 10 40 00 00 00 00 00 00 00 00 00", 128, 128, "004"), # MMC3, PRG 128, CHR ROM 128
0x21: ("10 00 40 00 00 00 00 00 00 00 00 00", 256,   0, "004"), # MMC3, PRG 256, CHR RAM
0x22: ("08 00 40 00 00 00 00 00 00 00 00 00", 128,   0, "004"), # MMC3, PRG 128, CHR RAM
0x84: ("02 08 30 00 00 00 00 00 00 00 00 00",  32,  64, "268"), # ?   , PRG  32, CHR ROM  64
0x44: ("02 04 30 00 00 00 00 00 00 00 00 00",  32,  32, "268"), # ?   , PRG  32, CHR ROM  32
0x54: ("02 02 30 00 00 00 00 00 00 00 00 00",  32,  16, "268"), # ?   , PRG  32, CHR ROM  16
0x64: ("02 01 00 00 00 00 00 00 00 00 00 00",  32,   8, "000"), # NROM, PRG  32, CHR ROM   8
0x73: ("01 01 00 00 00 00 00 00 00 00 00 00",  16,   8, "000"), # NROM, PRG  16, CHR ROM   8
}

typ_count = {}

if DEBUG:
    test_rom = bytearray(rom)

for igame in range(365):
    off = base + addrs_pos + igame * 4
    game_name_addr, game_index = unpack("<HH", rom[off:off+4])
    game_name_pos = game_name_addr - 0x8000 + base
    iend = rom.find(b"\xFF", game_name_pos)
    game_name = rom[game_name_pos:iend].decode()

    off = base + info_addrs_pos + game_index * 2
    game_info_addr = unpack("<H", rom[off:off+2])[0]
    game_info_pos = game_info_addr - 0x8000 + base
    # off
    # 00   : Маппер и размер
    # 01-02: Позиция PRG * 16KiB, LE
    # 03-04: Позиция CHR * 8KiB, LE
    # 05   : Mirroring
    # 06   : Всегда 00
    typ, prg_pos, chr_pos, mirr = unpack("<BHHB", rom[game_info_pos:game_info_pos+6])
    prg_pos *= 0x4000
    chr_pos *= 0x2000

    if DEBUG:
        info = rom[game_info_pos:game_info_pos+7]
        print(f"{game_name}\n    PRG pos {prg_pos:7X}, CHR pos {chr_pos:7X}, Mirr {mirr}, Raw {info.hex(' ')}")

    if typ in typ_count:
        typ_count[typ] += 1
    else:
        typ_count[typ] = 1

    if typ in headers:
        hdr, prg_size, chr_size, mapper = headers[typ]
        hdr = bytearray.fromhex("4E 45 53 1A" + hdr)
        if mirr == 0:
            hdr[6] |= 1
        if typ == 0xF2:
            if chr_pos != 0x27A0000:
                # BB Car
                hdr[5] |= 1
                chr_size = 8
        elif typ == 0x64:
            if chr_pos == 0x36F8000:
                # 3D Block, Pravila Dorozhnogo, Puteshestvie Po Europe
                chr_size = 0

        p = output_path / (f"{game_name} [m{mapper}].nes")
        p.write_bytes(hdr + rom[prg_pos:prg_pos+prg_size*1024] + rom[chr_pos:chr_pos+chr_size*1024])
        if DEBUG:
            test_rom[prg_pos:prg_pos+prg_size*1024] = bytes(prg_size*1024)
            test_rom[chr_pos:chr_pos+chr_size*1024] = bytes(chr_size*1024)

print("# Count")
for k, v in headers.items():
    print(f"{k:02X}, m{v[3]}, PRG {v[1]:3}, CHR {v[2]:3}: {typ_count[k]}")

if DEBUG:
    Path("test_rom.nes").write_bytes(bytes(0x10) + test_rom)
